#!/bin/bash

shellfu import jat

#
# Generate universally unique evidence ID
#
# uuevid__new() can be used to generate string that is
#
#  *  globally unique (random, much like UUID),
#  *  retained for duration of test,
#  *  reset with another test.
#
# Such value can be used for purposes of more reliable tracking.
# For example you can use it during setup as a "message" inside an
# e-mail, then by finding it in a message collected by client you
# *know* it was delivered.  Compared to arbitrary strings such as
# 'foo' or 'test', using this function results in proofs that
# cannot already appear by normal circumstances.
#
# A simple but realistic example:
#
#     jat__pstarts
#         {
#             echo "To: root@localhost"
#             echo "Subject: $(uuevid__new subj)"
#             echo
#             uuevid__new body
#         } | sendmail -t
#     jat__pend
#
#     jat__pstartt
#         jat__cmd -o mail.out mail <<<type
#         jat__cmd grep -w "Subject: $(uuevid subj)" mail.out
#         jat__cmd grep -we "$(uuevid body)" mail.out
#     jat__pend
#
# Note that the message in this example is guarranteed to be different
# in every test, so there'e no risk of picking up "noise" from
# previous test.  (Well, you should already take care of that in
# cleanup, but remember that SUT is buggy so that part might not be
# always reliable.)
#
# This is a bit similar to using $RANDOM, but gives a bit more power
# while removing the need to care about formatting, storage (no need
# to use globals) and in many cases will lead to more readable syntax.
#

#
# Main persistence root
#
UUEVID__ROOT=${UUEVID__ROOT:-/var/tmp/uuevid}

uuevid() {
    #
    # Print universally unique evidence ID named $1
    #
    # Usage:
    #
    #     uuevid NAME
    #
    # print value of UUEVID named NAME, previously generated by calling
    # uuevid__new().
    #
    local name                   # name of UUEVID to get
    local usage="usage: uuevid NAME"
    while true; do case $1 in
        -*) jat__log_error "$usage"; return 2 ;;
        *)  break
    esac done
    test -n
    name=$1
    test -n "$name" || {
        jat__log_error "$usage"
        return 2
    }
    __uuevid__validate "$name" || {
        jat__log_error "not a valid NAME: $name"
        return 2
    }
    file="$(__uuevid__tmpfile "$name")"
    test -e "$file" || {
        jat__log_error "no such UUEVID: $name"
        return 2
    }
    cat "$file"
}

uuevid__new() {
    #
    # Create universally unique evidence ID named $1
    #
    # Usage:
    #
    #     uuevid__new [-s SIZE] [-n] NAME
    #
    # generate and print UUEVID named NAME.  This generation function may
    # be called only once per NAME, the value can be retrieved by call to
    # uuevid().
    #
    # NAME must be composed of letter or underscore, followed by zero or
    # more digits, letters or underscores.
    #
    # By default, UUEVID has format of 'uuevid_XXXXXXXXXX', where 'X' is
    # a random character of `[:alnum:]` POSIX class under C locale.  SIZE
    # determines number of random characters ('X's) and is 10 by default.
    #
    # -n option changes the format to a decimal integer number, ie. no
    # `uuevid_` prefix (it's possible to get all-zeroes).
    #
    # Calling this function with same NAME twice generates error and does
    # not print anything; you can override the behavior by -f option, which
    # causes UUEVID to be re-generated silently.
    #
    local name                  # this UUEVID name
    local Fmt="uuevid_%s"       # UUEVID format
    local Size=10               # payload size
    local force=false           # silently overwrite same UUEVID
    local Class='[:alnum:]'     # POSIX class (actually 'tr' arg)
    local usage="usage: uuevid__new [-s SIZE] [-n] NAME"
    while true; do case $1 in
        -f) force=true; shift ;;
        -n) Class='0-9'; Fmt="%s"; shift ;;
        -s) Size=$2; shift 2 || { jat__log_error "$usage"; return 2; } ;;
        -*)  jat__log_error "$usage"; return 2 ;;
        *)  break
    esac done
    name=$1
    test -n "$name" || {
        jat__log_error "$usage"
        return 2
    }
    __uuevid__validate "$name" || {
        jat__log_error "not a valid NAME: $name"
        return 2
    }
    file="$(__uuevid__tmpfile "$name")"
    if test -e "$file"; then
        $force || {
            jat__log_error "UUEVID already exists: $name"
            return 3
        }
    fi
    __uuevid__mkuuevid >"$file"
    test -s "$file" || {
        jat__log_error "error creating UUEVID"
        return 3
    }
    jat__log_info "generated new UUEVID: $name='$(<"$file")'"
    cat "$file"
}

uuevid__reset() {
    #
    # Remove all UUEVIDs
    #
    rm -r "$UUEVID__ROOT/sessions"
}

uuevid__yaml() {
    #
    # Print all uuevids to stdout as YAML dict
    #
    local name  # each name in this session
    local pth   # session's name storage path
    pth="$UUEVID__ROOT/sessions/$__UUEVID__SESSID"
    find "$pth" -type f -printf '%P\n' \
      | while read -r name; do
            echo "$name: $(uuevid "$name")"
        done
}


#          #                            do not think ther is no evidence #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                             that you have crossed this line #

#
# Session ID
#
__UUEVID__SESSID=""

__uuevid__mkuuevid() {
    #
    # Print new UUEVID with size $Size, POSIX class $Class, format as $Fmt
    #
    local payload
    payload=$(</dev/urandom LC_ALL=C tr -dc "$Class" | head -c "$Size")
    test -n "$payload" || return 3
    #shellcheck disable=SC2059
    printf "$Fmt" "$payload"
}

__uuevid__tmpread() {
    #
    # Get result metadata from key $1
    #
    # See __uuevid__tmpfile() for key syntax.
    #
    local key=$1    # key to read
    cat "$(__uuevid__tmpfile "$key")"
}

__uuevid__tmpwrite() {
    #
    # Save result metadata value $2 under key $1
    #
    # See __uuevid__tmpfile() for key syntax.
    #
    local key=$1    # key to write
    local value=$2  # value to write
    local tgt       # target file path
    tgt=$(__uuevid__tmpfile "$key")
    echo "$value" > "$tgt"
}

__uuevid__tmpfile() {
    #
    # Dereference storage key $1
    #
    # Note: This function has a side effect within the storage structure that
    # directory for the data file is automatically created so that caller does
    # not need to.
    #
    local key=$1    # key to dereference
    local path      # final path
    path="$UUEVID__ROOT/sessions/$__UUEVID__SESSID/$key"
    mkdir -p "${path%/*}"
    echo "$path"
}


__uuevid__validate() {
    #
    # True if $1 is a valid UUEVID name or throw error
    #
    local str=$1
    grep -qx '[[:alpha:]_][[:alnum:]_]*' <<<"$str"
}

__shellfu_uuevid__init() {
    #
    # Initialize session based on $TESTID or UNIX date
    #
    __UUEVID__SESSID="unix$(date +%s)"
}

#shellfu module-version=0.0.5
